@@ -1,6 +1,6 @@
 /*
  * Parser for IDSgrep
- * Copyright (C) 2012  Matthew Skala
+ * Copyright (C) 2012, 2013  Matthew Skala
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -42,6 +42,35 @@ HASHED_STRING *close_bracket,*semicolon=NULL,*socked_head=NULL;
 		     || (((x)>='A') && ((x)<='F')))
 #define MYXVAL(x) (((x)&0x40)?(((x)&0xF)+9):((x)&0xF))
 
+int construct_utf8(int xval,char *ebuf) {
+   int clen;
+   
+   if (xval>=0x110000)
+     xval=0xFFFD;
+   if ((xval>=0xD800) && (xval<=0xDFFF))
+     xval=0xFFFD;
+   if (xval<0x80) {
+      clen=1;
+      ebuf[0]=xval;
+   } else if (xval<0x800) {
+      clen=2;
+      ebuf[0]=0xC0|(xval>>6);
+      ebuf[1]=0x80|(xval&0x3F);
+   } else if (xval<0x10000) {
+      clen=3;
+      ebuf[0]=0xE0|(xval>>12);
+      ebuf[1]=0x80|((xval>>6)&0x3F);
+      ebuf[2]=0x80|(xval&0x3F);
+   } else {
+      clen=4;
+      ebuf[0]=0xF0|(xval>>18);
+      ebuf[1]=0x80|((xval>>12)&0x3F);
+      ebuf[2]=0x80|((xval>>6)&0x3F);
+      ebuf[3]=0x80|(xval&0x3F);
+   }
+   return clen;
+}
+
 size_t parse(size_t len,char *inp) {
    int offs=0,clen,escaped,flag,xval,i;
    char ebuf[4],*eptr;
@@ -206,29 +235,7 @@ size_t parse(size_t len,char *inp) {
 		 continue;
 	    }
 	    eptr=ebuf;
-	    if (xval>=0x110000)
-	      xval=0xFFFD;
-	    if ((xval>=0xD800) && (xval<=0xDFFF))
-	      xval=0xFFFD;
-	    if (xval<0x80) {
-	       clen=1;
-	       ebuf[0]=xval;
-	    } else if (xval<0x800) {
-	       clen=2;
-	       ebuf[0]=0xC0|(xval>>6);
-	       ebuf[1]=0x80|(xval&0x3F);
-	    } else if (xval<0x10000) {
-	       clen=3;
-	       ebuf[0]=0xE0|(xval>>12);
-	       ebuf[1]=0x80|((xval>>6)&0x3F);
-	       ebuf[2]=0x80|(xval&0x3F);
-	    } else {
-	       clen=4;
-	       ebuf[0]=0xF0|(xval>>18);
-	       ebuf[1]=0x80|((xval>>12)&0x3F);
-	       ebuf[2]=0x80|((xval>>6)&0x3F);
-	       ebuf[3]=0x80|(xval&0x3F);
-	    }
+	    clen=construct_utf8(xval,ebuf);
 	    offs-=clen;
 	    break;
 
@@ -373,7 +380,20 @@ size_t parse(size_t len,char *inp) {
 	    /* put it in the string, then re-read it as closing bracket */
 	    close_bracket=hchar;
 	    parse_state=(PARSE_STATE)hchar->arity;
-	    memcpy(partstr,eptr,clen);
+	    switch (clen) {
+	     case 4: /* SNH */
+	       partstr[3]=eptr[3]; /* SNH */ /* no 4-byte chars are sugary */
+	       /* FALL THROUGH */
+	     case 3:
+	       partstr[2]=eptr[2];
+	       /* FALL THROUGH */
+	     case 2:
+	       partstr[1]=eptr[1];
+	       /* FALL THROUGH */
+	     case 1:
+	       partstr[0]=eptr[0];
+	       /* FALL THROUGH */
+	    }
 	    partstr_len=clen;
 	    delete_string(hchar);
 	    continue;
@@ -429,6 +449,7 @@ size_t parse(size_t len,char *inp) {
 		(parse_state==newstr->arity)) {
 	       tmps=newstr->canonical;
 	       delete_string(newstr);
+	       tmps->refs++;
 	       newstr=tmps;
 	    }
 	    
@@ -463,8 +484,26 @@ size_t parse(size_t len,char *inp) {
 	    }
 	    
 	    /* append the data */
-	    memcpy(partstr+partstr_len,eptr,clen);
-	    partstr_len+=clen;
+	    switch (clen) {
+	     case 4:
+	       partstr[partstr_len++]=*(eptr++);
+	       /* FALL THROUGH */
+	     case 3:
+	       partstr[partstr_len++]=*(eptr++);
+	       /* FALL THROUGH */
+	     case 2:
+	       partstr[partstr_len++]=*(eptr++);
+	       /* FALL THROUGH */
+	     case 1:
+	       partstr[partstr_len++]=*eptr;
+	       /* FALL THROUGH */
+	    }
+	    /* Note this is the code that the above switch() replaced, but
+	     * the switch() also harmlessly trashes eptr.  It works because
+	     * we know clen must be 1, 2, 3, or 4.
+	     *    memcpy(partstr+partstr_len,eptr,clen);
+	     *	  partstr_len+=clen;
+	     */
 	    offs+=clen;
 	 }
 	 
@@ -533,13 +572,11 @@ void register_brackets(char *opb,char *clb,int arity,int idx) {
    
    oph->arity=arity;
    oph->mate=clh;
-   oph->refs++;
    clh->mate=oph;
-   if (clh!=oph) clh->refs++;
    hashed_bracket[idx]=oph;
 }
 
-void register_special_functor(char *fctr,int arity,MATCH_FN mf) {
+void register_special_functor(char *fctr,int arity,MATCH_FN mf,BITVEC_FN bf) {
    HASHED_STRING *hs;
    
    hs=new_string(strlen(fctr),fctr);
@@ -551,6 +588,7 @@ void register_special_functor(char *fctr,int arity,MATCH_FN mf) {
 
    hs->arity=arity;
    hs->match_fn=mf;
+   hs->needle_bits_fn=bf;
 }
 
 void register_alias(char *fctr,char *canon) {
@@ -592,70 +630,85 @@ void register_syntax(void) {
    register_brackets("\xE3\x80\x94","\xE3\x80\x95",3,13); /* b tortoise */
    register_brackets("\xE3\x80\x98","\xE3\x80\x99",3,14); /* w tortoise */
    
-   register_special_functor(";",0,default_match_fn);
+   register_special_functor(";",0,default_match_fn,default_needle_fn);
 
-   register_special_functor(",",2,default_match_fn);
+   register_special_functor(",",2,default_match_fn,default_needle_fn);
 
-   register_special_functor("?",0,anything_match_fn);
+   register_special_functor("?",0,anything_match_fn,anything_needle_fn);
    register_alias("anything","?");
 
-   register_special_functor(".",1,anywhere_match_fn);
+   register_special_functor(".",1,anywhere_match_fn,anywhere_needle_fn);
    register_alias("anywhere",".");
 
-   register_special_functor("&",2,and_or_match_fn);
+   register_special_functor("&",2,and_or_match_fn,and_needle_fn);
    register_alias("and","&");  
 
-   register_special_functor("|",2,and_or_match_fn);
+   register_special_functor("|",2,and_or_match_fn,or_needle_fn);
    register_alias("or","|");
    
-   register_special_functor("!",1,not_match_fn);
+   register_special_functor("!",1,not_match_fn,not_needle_fn);
    register_alias("not","!");
 
-   register_special_functor("*",1,unord_match_fn);
+   register_special_functor("*",1,unord_match_fn,unord_needle_fn);
    register_alias("unord","*");
 
-   register_special_functor("=",1,equal_match_fn);
+   register_special_functor("=",1,equal_match_fn,equal_needle_fn);
    register_alias("equal","=");
 
-   register_special_functor("@",1,assoc_match_fn);
+   register_special_functor("@",1,assoc_match_fn,anything_needle_fn);
    register_alias("assoc","@");
 
-   register_special_functor("/",1,regex_match_fn);
+   register_special_functor("/",1,regex_match_fn,anything_needle_fn);
    register_alias("regex","/");
+   
+   register_special_functor("#",1,user_match_fn,anything_needle_fn);
+   register_alias("user","#");
 
-   register_special_functor("\xE2\xBF\xB0",2,default_match_fn);
+   register_special_functor("\xE2\xBF\xB0",2,
+			    default_match_fn,default_needle_fn);
    register_alias("lr","\xE2\xBF\xB0");
 
-   register_special_functor("\xE2\xBF\xB1",2,default_match_fn);
+   register_special_functor("\xE2\xBF\xB1",2,
+			    default_match_fn,default_needle_fn);
    register_alias("tb","\xE2\xBF\xB1");
 
-   register_special_functor("\xE2\xBF\xB2",3,default_match_fn);
+   register_special_functor("\xE2\xBF\xB2",3,
+			    default_match_fn,default_needle_fn);
    register_alias("lcr","\xE2\xBF\xB2");
 
-   register_special_functor("\xE2\xBF\xB3",3,default_match_fn);
+   register_special_functor("\xE2\xBF\xB3",3,
+			    default_match_fn,default_needle_fn);
    register_alias("tcb","\xE2\xBF\xB3");
 
-   register_special_functor("\xE2\xBF\xB4",2,default_match_fn);
+   register_special_functor("\xE2\xBF\xB4",2,
+			    default_match_fn,default_needle_fn);
    register_alias("enclose","\xE2\xBF\xB4");
 
-   register_special_functor("\xE2\xBF\xB5",2,default_match_fn);
+   register_special_functor("\xE2\xBF\xB5",2,
+			    default_match_fn,default_needle_fn);
    register_alias("wrapu","\xE2\xBF\xB5");
 
-   register_special_functor("\xE2\xBF\xB6",2,default_match_fn);
+   register_special_functor("\xE2\xBF\xB6",2,
+			    default_match_fn,default_needle_fn);
    register_alias("wrapd","\xE2\xBF\xB6");
 
-   register_special_functor("\xE2\xBF\xB7",2,default_match_fn);
+   register_special_functor("\xE2\xBF\xB7",2,
+			    default_match_fn,default_needle_fn);
    register_alias("wrapl","\xE2\xBF\xB7");
 
-   register_special_functor("\xE2\xBF\xB8",2,default_match_fn);
+   register_special_functor("\xE2\xBF\xB8",2,
+			    default_match_fn,default_needle_fn);
    register_alias("wrapul","\xE2\xBF\xB8");
 
-   register_special_functor("\xE2\xBF\xB9",2,default_match_fn);
+   register_special_functor("\xE2\xBF\xB9",2,
+			    default_match_fn,default_needle_fn);
    register_alias("wrapur","\xE2\xBF\xB9");
 
-   register_special_functor("\xE2\xBF\xBA",2,default_match_fn);
+   register_special_functor("\xE2\xBF\xBA",2,
+			    default_match_fn,default_needle_fn);
    register_alias("wrapll","\xE2\xBF\xBA");
 
-   register_special_functor("\xE2\xBF\xBB",2,default_match_fn);
+   register_special_functor("\xE2\xBF\xBB",2,
+			    default_match_fn,default_needle_fn);
    register_alias("overlap","\xE2\xBF\xBB");
 }
